# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 6.6.2
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore

qt_resource_data = b"\
\x00\x00\x02\x04\
\x89\
PNG\x0d\x0a\x1a\x0a\x00\x00\x00\x0dIHDR\x00\
\x00\x00:\x00\x00\x00\x09\x08\x06\x00\x00\x00cU?+\
\x00\x00\x00\x09pHYs\x00\x00-\xd0\x00\x00-\xd0\
\x01\xe7\x9cB{\x00\x00\x00\x19tEXtSof\
tware\x00www.inksca\
pe.org\x9b\xee<\x1a\x00\x00\x01\x91ID\
ATH\x89\xe5\x95AK\x15Q\x18\x86\x9f\xebU\x17\
\xee\x14\x83(I7!BP\x7f \x8c \xbd\xb5J\
\xfc\x01\xf6\x03\xf2\xbf\xa9\x08\xa9a\x7f\xa1\x22\x02sg\
\xdd\x8c\xa8\xae!z7.\x9e\x163\xa3\xe7\x8e\xdf\x19\
\x09w\xf5\xc2,\xe6=\xcfy9g\xce\xf7\x9di\xa9\
\xfc\x0f\x1a\x0e\xbc\x19`\x018\x03\xb6\x80o\x0d\xf3\x87\
\x80\x0ep\x17\xf8\x08\xec\x02M_n\xaa\xe4\x05\xb6\x81\
n\x03\xdb\x02\x9e\x00s\xc0\xa7\x92o\xca\xbe\x09<\xa3\
\xd8\xd36p00\xaa\xa6\xcf\xb2\xda\xf7BG\xea|\
\x8d\xa9\x9eQu\xd7Am\xa8\xed\x0c\xffT=I\xd8\
cu1\xc3\xb6\xd5\xcdZ\xf6ku$\xc3?R\x7f\
'l_]J\x99\x14\x1eV{^\xd6^&|5\
`UW\x02\xb6\xa5~\x0d\xd8n9V\xe7_d\xb2\
_f\xd6\xb2\x1f\xb0\xbf\xca=\xa12\x94\x1c\xee40\
\x1e\x94\xc4,0\x16\xf8\xf73%\xf4 \xf0n\x00\xb7\
\x02\xff60y\xcd\xec1\x8a\xd6\xa9k\x02\xb8S\xbd\
\xa4\x1b=\x00z\xc1\x84=\xa0\x1f\xf8\xef2\x8by\x1b\
x?\x80\xc3\xc0\xef\x02?\x03\xff\xfd_d\xf7\x81\xfd\
\xc0\xef\x01\x9f\xcf\xdfj%\xb0\xa4\x9e&\xc7\x7f\xa4>\
\xcc\x94\xcb\xa8E\xdf\xa4Z3\xdf\xa3\x1d\x8b\xbe\xact\
\xac.d\xd8\xb6\xba^\xcb\xde1\xdf\xa3\xf3\xe5Z+\
\x9d\xaa\xcfS\xa6\xe5\xe5\xdf\xcb4\x17\xb7\xee+\xe0{\
\xe6\xebBq3.R\x94\xf7\x07\xe0M\x03\x0bE\xa9\
V\xb7\xee\x16\xf1)\xa7\xd9\x8f\x81{\x14U\xb5\xc3\xd5\
\xb7n\x07\x18)\xb3\xbf\x0c\x84\x05\x1b\xfd'\xf5\x07\xe1\
#\x85k\x01\x84\xa6V\x00\x00\x00\x00IEND\xae\
B`\x82\
"

qt_resource_name = b"\
\x00\x08\
\x0f\xa5\xa3\xf9\
\x00v\
\x00i\x00c\x00o\x00d\x00e\x00p\x00y\
\x00\x06\
\x06\x8a\x9c\xb3\
\x00a\
\x00s\x00s\x00e\x00t\x00s\
\x00\x0a\
\x0b-\x87\xc7\
\x00h\
\x00a\x00n\x00d\x00l\x00e\x00.\x00p\x00n\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x16\x00\x02\x00\x00\x00\x01\x00\x00\x00\x03\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00(\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x90U\x9b\xbc_\
"


def qInitResources():
    QtCore.qRegisterResourceData(
        0x03, qt_resource_struct, qt_resource_name, qt_resource_data
    )


def qCleanupResources():
    QtCore.qUnregisterResourceData(
        0x03, qt_resource_struct, qt_resource_name, qt_resource_data
    )


qInitResources()
