# ViCodePy - A video coder for Experimental Psychology
#
# Copyright (C) 2024 Esteban Milleret
# Copyright (C) 2024 Rafael Laboissière
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

import pandas as pd


class Data:
    def __init__(self, filename, file_format, current_format):
        self.data = pd.read_csv(filename, sep=None, engine="python")
        self.update_format(file_format, current_format)

    def update_format(self, file_format, current_format):
        for fmt in range(file_format + 1, current_format + 1):
            if fmt == 2:
                pass
            if fmt == 3:
                self.data.rename(columns={"label": "event"}, inplace=True)
